@@ -18,11 +18,10 @@
 Create and parse torrent files and magnet URIs
 """
 
-__version__ = '3.1.1'
+__version__ = '4.0.1'
 
 from ._errors import *
 from ._magnet import Magnet
+from ._stream import TorrentFileStream
 from ._torrent import Torrent
-# Export File class so users can add items to Torrent.files and Filepath so it
-# is documented by readthedocs.
 from ._utils import File, Filepath
