@@ -30,6 +30,7 @@ import flatbencode as bencode
 from . import __version__
 from . import _errors as error
 from . import _generate as generate
+from . import _reuse as reuse
 from . import _utils as utils
 
 _PACKAGE_NAME = __name__.split('.')[0]
@@ -92,7 +93,7 @@ class Torrent():
     >>> t = Torrent.read('my_torrent.torrent')
     >>> t.comment
     "This is my first torrent. Let's rock!"
-    >>> t.metainfo['info']['more stuff']
+    >>> t.metainfo['more stuff']
     {'bar': ['x', 'y', 'z'], 'foo': 12}
     """
 
@@ -156,7 +157,7 @@ class Torrent():
         :attr:`include_regexs`.
 
         Setting or manipulating this property updates
-        :attr:`metainfo`\ ``['info']``:
+        :attr:`metainfo`\\ ``['info']``:
 
         - ``name``, ``piece length`` and ``files`` or ``length`` are set.
         - ``pieces`` and ``md5sum`` are removed if they exist.
@@ -178,6 +179,14 @@ class Torrent():
                               for fp in utils.list_files(basepath))
             self._set_files(filepaths, basepath)
 
+    @property
+    def location(self):
+        """
+        :attr:`path` without :attr:`name` or `None` if :attr:`path` is `None`
+        """
+        if self.path:
+            return self._path.parent
+
     @property
     def files(self):
         """
@@ -187,7 +196,7 @@ class Torrent():
         deduplicated.  Every path starts with :attr:`name`.
 
         Setting or manipulating this property updates
-        :attr:`metainfo`\ ``['info']``:
+        :attr:`metainfo`\\ ``['info']``:
 
         - ``name``, ``piece length`` and ``files`` or ``length`` are set.
         - ``pieces`` and ``md5sum`` are removed if they exist.
@@ -246,7 +255,7 @@ class Torrent():
         deduplicated.  Directories are resolved into a list of files.
 
         Setting or manipulating this property updates
-        :attr:`metainfo`\ ``['info']``:
+        :attr:`metainfo`\\ ``['info']``:
 
         - ``name``, ``piece length`` and ``files`` or ``length`` are set.
         - ``pieces`` and ``md5sum`` are removed if they exist.
@@ -292,7 +301,7 @@ class Torrent():
     def _set_files(self, files, basepath=None):
         """
         Update ``name`` and ``files`` or ``length``, remove ``pieces`` and
-        ``md5sum`` in :attr:`metainfo`\ ``['info']``
+        ``md5sum`` in :attr:`metainfo`\\ ``['info']``
 
         :param files: Sequence of :class:`File`
         :param basepath: path-like that all paths in `files` start with; may be
@@ -385,7 +394,7 @@ class Torrent():
         ======== ============================
         Wildcard Description
         ======== ============================
-        \*       matches everything
+        \\*      matches everything
         ?        matches any single character
         [SEQ]    matches any character in SEQ
         [!SEQ]   matches any char not in SEQ
@@ -503,7 +512,7 @@ class Torrent():
         is set to the default.
 
         Setting this property sets or removes ``name`` in
-        :attr:`metainfo`\ ``['info']``.
+        :attr:`metainfo`\\ ``['info']``.
         """
         if 'name' not in self.metainfo['info'] and self.path is not None:
             self.metainfo['info']['name'] = self.path.name
@@ -580,13 +589,13 @@ class Torrent():
         Length of each piece in bytes
 
         If set to ``None`` and :attr:`size` is larger than 0, use the return
-        value of :attr:`calculate_piece_size`.
+        value of :attr:`calculate_piece_size`.  If set to ``None`` and
+        :attr:`size` is smaller than 1, remove
+        :attr:`metainfo`\\ ``['info']``\\ ``['piece length']`` .
 
         Setting this property sets or removes ``piece length`` in
-        :attr:`metainfo`\ ``['info']``.
+        :attr:`metainfo`\\ ``['info']``.
         """
-        if 'piece length' not in self.metainfo['info']:
-            self.piece_size = None  # Calculate piece size
         return self.metainfo['info'].get('piece length', 0)
 
     @piece_size.setter
@@ -665,11 +674,10 @@ class Torrent():
 
     @property
     def pieces(self):
-        """
-        Number of pieces the content is split into
-        """
-        if self.size > 0:
-            return math.ceil(self.size / self.piece_size)
+        """Number of pieces the content is split into"""
+        size, piece_size = self.size, self.piece_size
+        if size and piece_size and size > 0 and piece_size > 0:
+            return math.ceil(size / piece_size)
         else:
             return 0
 
@@ -695,20 +703,20 @@ class Torrent():
         validation and deduplication.  You can set this property to a URL, an
         iterable of URLs or an iterable of iterables of URLs (i.e. "tiers").
 
-        This property automatically sets :attr:`metainfo`\ ``['announce']`` and
-        :attr:`metainfo`\ ``['announce-list']`` when it is manipulated or set
+        This property automatically sets :attr:`metainfo`\\ ``['announce']`` and
+        :attr:`metainfo`\\ ``['announce-list']`` when it is manipulated or set
         according to these rules:
 
-        - If it contains a single URL, :attr:`metainfo`\ ``['announce']`` is set
-          and :attr:`metainfo`\ ``['announce-list']`` is removed if it exists.
+        - If it contains a single URL, :attr:`metainfo`\\ ``['announce']`` is set
+          and :attr:`metainfo`\\ ``['announce-list']`` is removed if it exists.
 
-        - If it contains an iterable of URLs, :attr:`metainfo`\ ``['announce']``
-          is set to the first URL and :attr:`metainfo`\ ``['announce-list']`` is
+        - If it contains an iterable of URLs, :attr:`metainfo`\\ ``['announce']``
+          is set to the first URL and :attr:`metainfo`\\ ``['announce-list']`` is
           set to a list of tiers, one tier for each URL.
 
         - If it contains an iterable of iterables of URLs,
-          :attr:`metainfo`\ ``['announce']`` is set to the first URL of the first
-          iterable and :attr:`metainfo`\ ``['announce-list']`` is set to a list
+          :attr:`metainfo`\\ ``['announce']`` is set to the first URL of the first
+          iterable and :attr:`metainfo`\\ ``['announce-list']`` is set to a list
           of tiers, one tier for each iterable of URLs.
 
         :raises URLError: if any of the announce URLs is invalid
@@ -756,8 +764,8 @@ class Torrent():
         http://bittorrent.org/beps/bep_0019.html
 
         The list returned by this property automatically updates
-        :attr:`metainfo`\ ``['url-list']`` when manipulated.  Setting this
-        property sets :attr:`metainfo`\ ``['url-list']``.
+        :attr:`metainfo`\\ ``['url-list']`` when manipulated.  Setting this
+        property sets :attr:`metainfo`\\ ``['url-list']``.
 
         :raises URLError: if any URL is invalid
         :raises ValueError: if set to anything that isn't an iterable or a
@@ -792,8 +800,8 @@ class Torrent():
         http://bittorrent.org/beps/bep_0017.html
 
         The list returned by this property automatically updates
-        :attr:`metainfo`\ ``['httpseeds']`` when manipulated.  Setting this
-        property sets :attr:`metainfo`\ ``['httpseeds']``.
+        :attr:`metainfo`\\ ``['httpseeds']`` when manipulated.  Setting this
+        property sets :attr:`metainfo`\\ ``['httpseeds']``.
 
         :raises URLError: if any URL is invalid
         :raises ValueError: if set to anything that isn't an iterable or a
@@ -826,7 +834,7 @@ class Torrent():
         Whether torrent should use trackers exclusively for peer discovery
 
         ``True`` or ``False`` if
-        :attr:`metainfo`\ ``['info']``\ ``['private']`` exists, ``None``
+        :attr:`metainfo`\\ ``['info']``\\ ``['private']`` exists, ``None``
         otherwise.
         """
         if 'private' in self.metainfo['info']:
@@ -846,7 +854,7 @@ class Torrent():
         """
         Comment string or ``None`` for no comment
 
-        Setting this property sets or removes :attr:`metainfo`\ ``['comment']``.
+        Setting this property sets or removes :attr:`metainfo`\\ ``['comment']``.
         """
         return self.metainfo.get('comment', None)
 
@@ -866,7 +874,7 @@ class Torrent():
         :meth:`datetime.datetime.fromtimestamp`.
 
         Setting this property sets or removes
-        :attr:`metainfo`\ ``['creation date']``.
+        :attr:`metainfo`\\ ``['creation date']``.
         """
         return self.metainfo.get('creation date', None)
 
@@ -887,7 +895,7 @@ class Torrent():
         Application name or ``None`` for no creator
 
         Setting this property sets or removes
-        :attr:`metainfo`\ ``['created by']``.
+        :attr:`metainfo`\\ ``['created by']``.
         """
         return self.metainfo.get('created by', None)
 
@@ -904,7 +912,7 @@ class Torrent():
         Source string or ``None`` for no source
 
         Setting this property sets or removes
-        :attr:`metainfo`\ ``['info']``\ ``['created by']``.
+        :attr:`metainfo`\\ ``['info']``\\ ``['created by']``.
         """
         return self.metainfo['info'].get('source', None)
 
@@ -953,7 +961,7 @@ class Torrent():
         This allows for cross-seeding without changing :attr:`piece_size`.
 
         Setting this property to ``True`` sets
-        :attr:`metainfo`\ ``['info']``\ ``['entropy']`` to a random integer.
+        :attr:`metainfo`\\ ``['info']``\\ ``['entropy']`` to a random integer.
         Setting it to ``False`` removes that field.
         """
         return bool(self.metainfo['info'].get('entropy', False))
@@ -964,7 +972,7 @@ class Torrent():
             # According to BEP0003 "Integers have no size limitation", but some
             # parsers seem to have problems with large numbers.
             import random
-            self.metainfo['info']['entropy'] = random.randint(-2e9, 2e9)
+            self.metainfo['info']['entropy'] = random.randint(int(-2e9), int(2e9))
         else:
             self.metainfo['info'].pop('entropy', None)
 
@@ -982,7 +990,7 @@ class Torrent():
         """
         Hash pieces and report progress to `callback`
 
-        This method sets :attr:`metainfo`\ ``['info']``\ ``['pieces']`` after
+        This method sets :attr:`metainfo`\\ ``['info']``\\ ``['pieces']`` after
         all pieces are hashed successfully.
 
         :param int threads: How many threads to use for hashing pieces or
@@ -1009,91 +1017,153 @@ class Torrent():
         :return: ``True`` if all pieces were successfully hashed, ``False``
             otherwise
         """
-        self.metainfo['info']['pieces'] = bytes()
-        filepaths = self.filepaths
-
         if self.path is None:
             raise RuntimeError('generate() called with no path specified')
-        elif sum(utils.real_size(fp) for fp in filepaths) < 1:
+        elif sum(utils.real_size(fp) for fp in self.filepaths) < 1:
             raise error.PathError(self.path, msg='Empty or all files excluded')
 
-        if callback is not None:
-            maybe_cancel = generate.CancelCallback(callback, interval)
-        else:
-            maybe_cancel = None
-        threads = threads or NCORES
+        hasher_threads = threads or NCORES
 
-        # Read piece_size'd chunks from disk and push them to queue for hashing
-        reader = generate.Reader(filepaths=filepaths,
-                                 piece_size=self.piece_size,
-                                 queue_size=threads * 3)
+        # Read piece_size'd chunks from disk and send them to HasherPool
+        reader = generate.Reader(torrent=self, queue_size=hasher_threads * 3)
 
-        # Pool of workers that pull from reader's piece queue, calculate the
-        # hashes, and quickly offload the results to a hash queue
-        hasher_threadpool = generate.HasherPool(threads, reader.piece_queue)
+        # Multiple threads that get chunks from Reader, calculate the hashes,
+        # and push them to a hash queue
+        hashers = generate.HasherPool(
+            hasher_threads=hasher_threads,
+            piece_queue=reader.piece_queue,
+        )
 
-        # Pull from the hash queue and call status/cancel callback
-        def collector_callback(filepath, pieces_done, piece_index, piece_hash, exc,
-                               maybe_cancel=maybe_cancel, torrent=self, pieces_total=self.pieces):
-            if exc is not None:
-                raise exc
-            elif maybe_cancel is not None:
-                maybe_cancel(cb_args=(torrent, filepath, pieces_done, pieces_total),
-                             # Always call callback after the last piece was hashed
-                             force_call=pieces_done >= pieces_total)
-        collector_thread = generate.Collector(hasher_threadpool.hash_queue,
-                                              callback=collector_callback)
+        # Collect piece hashes from HasherPool and call `callback` for status
+        # reporting/cancellation
+        collector = generate.Collector(
+            torrent=self,
+            reader=reader,
+            hashers=hashers,
+            callback=generate.GenerateCallback(
+                callback=callback,
+                interval=interval,
+                torrent=self,
+            ),
+        )
 
-        if maybe_cancel:
-            maybe_cancel.on_cancel(reader.stop,
-                                   hasher_threadpool.stop,
-                                   collector_thread.stop)
-
-        try:
-            reader.read()
-        except BaseException:
-            hasher_threadpool.stop()
-            collector_thread.stop()
-            raise
-        finally:
-            hasher_threadpool.join()
-            collector_thread.join()
-
-        # Store generated hashes in metainfo
-        hashes_count = len(collector_thread.hashes) / 20
+        # Collect piece hashes
+        piece_hashes = collector.collect()
+        concatenated_piece_hashes = b''.join(piece_hashes)
+        hashes_count = len(concatenated_piece_hashes) / 20
         if hashes_count == self.pieces:
-            self.metainfo['info']['pieces'] = collector_thread.hashes
+            self.metainfo['info']['pieces'] = concatenated_piece_hashes
             return True
         elif hashes_count < self.pieces:
-            # Hashing was aborted by callback
+            # Hashing was cancelled
             return False
         else:
             raise RuntimeError('Unexpected number of hashes generated: '
                                f'{hashes_count} instead of {self.pieces}')
 
-    def _verify_prepare(self, path, callback, interval):
-        """Common tasks of :meth:`verify` and :meth:`verify_filesize`"""
+    def verify(self, path, threads=None, callback=None, interval=0):
+        """
+        Check if `path` contains all the data specified in this torrent
+
+        Generate hashes from file content and compare them to the ones stored in
+        :attr:`metainfo`\\ ``['info']``\\ ``['pieces']``.
+
+        :param str path: Directory or file to read from
+        :param int threads: How many threads to use for hashing pieces or
+            ``None`` to use one thread per available CPU core
+        :param callable callback: Callable to report progress and/or abort
+
+            `callback` must accept 7 positional arguments:
+
+                1. Torrent instance (:class:`Torrent`)
+                2. File path in file system (:class:`str`)
+                3. Number of checked pieces (:class:`int`)
+                4. Total number of pieces (:class:`int`)
+                5. Index of the current piece (:class:`int`)
+                6. SHA1 hash of the current piece (:class:`bytes` or ``None``)
+                7. Exception (:class:`TorfError` or ``None``)
+
+            If `callback` returns anything that is not ``None``, verification is
+            stopped.
+
+        :param float interval: Minimum number of seconds between calls to
+            `callback` (if 0, `callback` is called once per piece); this is
+            ignored if an error is found
+
+        If a callback is specified, exceptions are not raised but passed to
+        `callback` instead.
+
+        :raises VerifyContentError: if a file contains unexpected data
+        :raises VerifyIsDirectoryError: if `path` is a directory and this
+            torrent contains a single file
+        :raises VerifyNotDirectoryError: if `path` is a file and this torrent
+            contains a directory
+        :raises ReadError: if a file is not readable
+        :raises MetainfoError: if :meth:`validate` fails
+
+        :return: ``True`` if `path` is verified successfully, ``False``
+            otherwise
+        """
+        # First make sure we are a valid torrent
         self.validate()
 
-        if callback is not None:
-            callback = generate.CancelCallback(callback, interval=interval)
+        # Wrapper around callback function that compares hashes
+        verify_callback = generate.VerifyCallback(
+            callback=callback,
+            interval=interval,
+            torrent=self,
+            path=path,
+        )
+
+        def early_exception(exception):
+            piece_index = 0
+            pieces_done = 0
+            pieces_total = self.pieces
+            filepath = None
+            piece_hash = None
+            exceptions = (exception,)
+            verify_callback(piece_index, pieces_done, pieces_total, filepath, piece_hash, exceptions)
+
+        if self.mode == 'singlefile' and os.path.isdir(path):
+            early_exception(error.VerifyIsDirectoryError(path))
+            return False
+
+        elif self.mode == 'multifile' and not os.path.isdir(path):
+            early_exception(error.VerifyNotDirectoryError(path))
+            return False
+
         else:
-            # Stop the verification process if there was an exception
-            callback = generate.CancelCallback(lambda *cb_args: cb_args[-1])
+            hasher_threads = threads or NCORES
 
-        # Generate an ordered list of file system paths and their corresponding
-        # paths inside the torrent
-        paths = []
-        for torrent_filepath in self.files:
-            torrent_subpath = torrent_filepath.parts[1:]
-            fs_filepath = pathlib.Path(path, *torrent_subpath)
-            paths.append((fs_filepath, torrent_filepath))
+            # Read piece_size'd chunks from disk and send them to HasherPool
+            reader = generate.Reader(
+                torrent=self,
+                queue_size=hasher_threads * 3,
+                path=path,
+            )
 
-        return paths, callback
+            # Multiple threads that get chunks from Reader, calculate the hashes,
+            # and push them to a hash queue
+            hashers = generate.HasherPool(
+                hasher_threads=hasher_threads,
+                piece_queue=reader.piece_queue,
+            )
+
+            # Collect piece hashes from HasherPool and call `callback` for status
+            # reporting/cancellation
+            collector = generate.Collector(
+                torrent=self,
+                reader=reader,
+                hashers=hashers,
+                callback=verify_callback,
+            )
+
+            piece_hashes = collector.collect()
+            return piece_hashes == self.hashes
 
     def verify_filesize(self, path, callback=None):
         """
-        Check if `path` looks like it should contain all the data of this torrent
+        Check if `path` has the expected file size
 
         Walk through :attr:`files` and check if each file exists relative to
         `path`, is readable and has the correct size.  Excess files in `path`
@@ -1121,7 +1191,7 @@ class Torrent():
         `callback` instead.
 
         :raises VerifyFileSizeError: if a file has an unexpected size
-        :raises VerifyNotDirectoryError: if `path` is a directory and this
+        :raises VerifyIsDirectoryError: if `path` is a directory and this
             torrent contains a single file
         :raises ReadError: if any file's size can't be determined
         :raises MetainfoError: if :meth:`validate` fails
@@ -1129,32 +1199,55 @@ class Torrent():
         :return: ``True`` if `path` is verified successfully, ``False``
             otherwise
         """
-        raise_exceptions = not callback
-        filepaths, cancel = self._verify_prepare(path, callback, interval=0)
+        # First make sure we are a valid torrent
+        self.validate()
+
+        # Generate an ordered list of file system paths and their corresponding
+        # paths inside the torrent
+        # NOTE: The last segment in `path` is supposed to be the torrent name so
+        #       we must remove the name from each `file`.  This allows
+        #       verification of any renamed file/directory against a torrent.
+        filepaths = tuple(
+            (
+                utils.File([path, *file.parts[1:]], file.size),
+                file,
+            )
+            for file in self.files
+        )
         files_total = len(filepaths)
+
+        def cancel(file_index, exception):
+            if callback:
+                fs_filepath = filepaths[file_index][0]
+                torrent_filepath = filepaths[file_index][1]
+                files_done = file_index + 1
+                return_value = callback(self, fs_filepath, torrent_filepath,
+                                        files_done, files_total, exception)
+                if return_value is not None:
+                    return True
+            elif exception:
+                # Raise exception if there is no callback to handle it
+                raise exception
+            else:
+                return False
+
         exception = None
 
-        for files_done,(fs_filepath,torrent_filepath) in enumerate(filepaths, start=1):
+        # If we expect a file, check if path is a file.  We don't need to check
+        # for a directory if we expect one because we are iterating over files
+        # (filepaths), so the path "foo/bar/baz" will result in a ReadError if
+        # "foo" or "foo/bar" is a file.
+        if self.mode == 'singlefile' and os.path.isdir(path):
+            exception = error.VerifyIsDirectoryError(path)
+            cancel(file_index=0, exception=exception)
+            return False
+
+        for file_index, (fs_filepath, torrent_filepath) in enumerate(filepaths):
             # Check if path exists
             if not os.path.exists(fs_filepath):
                 exception = error.ReadError(errno.ENOENT, fs_filepath)
-                if cancel(cb_args=(self, fs_filepath, torrent_filepath,
-                                   files_done, files_total, exception),
-                          force_call=True):
-                    break
-                else:
-                    continue
-
-            # If we expect a file, check if path is a file.  We don't need to
-            # check for a directory if we expect one because we are iterating
-            # over files (filepaths), so the path "foo/bar/baz" will result in a
-            # ReadError if "foo" or "foo/bar" is a file.
-            if self.mode == 'singlefile' and os.path.isdir(path):
-                exception = error.VerifyNotDirectoryError(fs_filepath)
-                if cancel(cb_args=(self, fs_filepath, torrent_filepath,
-                                   files_done, files_total, exception),
-                          force_call=True):
-                    break
+                if cancel(file_index, exception):
+                    return False
                 else:
                     continue
 
@@ -1163,148 +1256,19 @@ class Torrent():
             expected_size = self.partial_size(torrent_filepath)
             if fs_filepath_size != expected_size:
                 exception = error.VerifyFileSizeError(fs_filepath, fs_filepath_size, expected_size)
-                if cancel(cb_args=(self, fs_filepath, torrent_filepath,
-                                   files_done, files_total, exception),
-                          force_call=True):
-                    break
+                if cancel(file_index, exception):
+                    return False
                 else:
                     continue
 
             # Report no error for current file
-            if cancel(cb_args=(self, fs_filepath, torrent_filepath, files_done, files_total, None),
-                      force_call=True):
-                break
+            if cancel(file_index, exception=None):
+                return False
 
         if exception:
-            if raise_exceptions:
-                raise exception
-            else:
-                return False
-        else:
-            return True
-
-    def verify(self, path, skip_on_error=True, threads=None, callback=None, interval=0):
-        """
-        Check if `path` contains all the data specified in this torrent
-
-        Generate hashes from the contents of :attr:`files` and compare them to
-        the ones stored in :attr:`metainfo`\ ``['info']``\ ``['pieces']``.
-
-        :param str path: Directory or file to check
-        :param bool skip_on_error: Whether to stop hashing pieces from file if a
-            piece from it is corrupt
-        :param int threads: How many threads to use for hashing pieces or
-            ``None`` to use one thread per available CPU core
-        :param callable callback: Callable to report progress and/or abort
-
-            `callback` must accept 7 positional arguments:
-
-                1. Torrent instance (:class:`Torrent`)
-                2. File path in file system (:class:`str`)
-                3. Number of checked pieces (:class:`int`)
-                4. Total number of pieces (:class:`int`)
-                5. Index of the current piece (:class:`int`)
-                6. SHA1 hash of the current piece or ``None`` (:class:`bytes`)
-                7. Exception (:class:`TorfError`) or ``None``
-
-            If `callback` returns anything that is not ``None``, verification is
-            stopped.
-
-        :param float interval: Minimum number of seconds between calls to
-            `callback` (if 0, `callback` is called once per piece); this is
-            ignored if an error is found
-
-        If a callback is specified, exceptions are not raised but passed to
-        `callback` instead.
-
-        :raises VerifyContentError: if a file contains unexpected data
-        :raises ReadError: if a file is not readable
-        :raises MetainfoError: if :meth:`validate` fails
-
-        :return: ``True`` if `path` is verified successfully, ``False``
-            otherwise
-        """
-        raise_exceptions = not callback
-        filepaths, maybe_cancel = self._verify_prepare(path, callback, interval=interval)
-        fs_filepaths = tuple(x[0] for x in filepaths)
-
-        if self.mode == 'singlefile' and os.path.isdir(path):
-            exception = error.VerifyNotDirectoryError(path)
-            maybe_cancel(cb_args=(self, fs_filepaths[0], 0, self.pieces, 0, None, exception),
-                         force_call=True)
-        elif self.mode == 'multifile' and not os.path.isdir(path):
-            exception = error.VerifyIsDirectoryError(path)
-            maybe_cancel(cb_args=(self, fs_filepaths[0], 0, self.pieces, 0, None, exception),
-                         force_call=True)
-        else:
-            exception = None
-            threads = threads or NCORES
-            reader = generate.Reader(filepaths=fs_filepaths,
-                                     file_sizes={fs_path:self.partial_size(t_path)
-                                                 for fs_path,t_path in filepaths},
-                                     piece_size=self.piece_size,
-                                     queue_size=threads * 3,
-                                     skip_on_error=skip_on_error)
-
-            # Pool of workers that pull from reader_thread's piece queue, calculate
-            # the hashes, and quickly offload the results to a hash queue
-            hasher_threadpool = generate.HasherPool(threads, reader.piece_queue,
-                                                    file_was_skipped=reader.file_was_skipped)
-
-            # Pull from the hash queue; also call `callback` and maybe stop everything
-            def collector_callback(filepath, pieces_done, piece_index, piece_hash, exc,
-                                   torrent=self, pieces_total=self.pieces, piece_size=self.piece_size,
-                                   exp_hashes=self.hashes,
-                                   exp_filesizes={fs_path : self.partial_size(t_path)
-                                                  for fs_path,t_path in filepaths}):
-                nonlocal exception
-                if exc is not None:
-                    # Reader raised exception
-                    exception = exc
-                    maybe_cancel(cb_args=(torrent, filepath, pieces_done, pieces_total,
-                                          piece_index, piece_hash, exception),
-                                 force_call=True)
-                elif piece_hash is not None and piece_hash != exp_hashes[piece_index]:
-                    # Piece hash doesn't match
-                    if skip_on_error:
-                        reader.skip_file(filepath, piece_index)
-                    file_sizes = tuple((fs_path, self.partial_size(t_path))
-                                       for fs_path,t_path in filepaths)
-                    exception = error.VerifyContentError(piece_index, piece_size, file_sizes)
-                    maybe_cancel(cb_args=(torrent, filepath, pieces_done, pieces_total,
-                                          piece_index, piece_hash, exception),
-                                 force_call=True)
-                else:
-                    # No error; report progress
-                    maybe_cancel(cb_args=(torrent, filepath, pieces_done, pieces_total,
-                                          piece_index, piece_hash, None),
-                                 # Always call callback after the last piece was
-                                 # hashed to report completion
-                                 force_call=pieces_done >= pieces_total)
-            collector_thread = generate.Collector(hasher_threadpool.hash_queue,
-                                                  callback=collector_callback,
-                                                  file_was_skipped=reader.file_was_skipped)
-            maybe_cancel.on_cancel(reader.stop,
-                                   hasher_threadpool.stop,
-                                   collector_thread.stop)
-            try:
-                reader.read()
-            except BaseException:
-                hasher_threadpool.stop()
-                collector_thread.stop()
-                raise
-            finally:
-                hasher_threadpool.join()
-                collector_thread.join()
-                if collector_thread.exception is not None:
-                    exception = collector_thread.exception
-
-        # Raise exception unless it was passed to callback
-        if exception:
-            if raise_exceptions:
-                raise exception
-            else:
-                return False
+            # `exception` is just an indicator of success/failure. At this point
+            # it was already raised or passed to cancel().
+            return False
         else:
             return True
 
@@ -1564,7 +1528,7 @@ class Torrent():
                 metainfo = utils.decode_dict(metainfo_enc)
 
             # "info" must be a dictionary.  If validation is not wanted, it's OK
-            # if it doesn't exist because the "metainfo" property will add
+            # if it doesn't exist because the "metainfo" property will add it
             # automatically.
             utils.assert_type(metainfo, ('info',), (dict,), must_exist=validate)
 
@@ -1615,6 +1579,114 @@ class Torrent():
         cp._metainfo = deepcopy(self._metainfo)
         return cp
 
+    def reuse(self, path, callback=None, interval=0):
+        """
+        Copy ``pieces`` and ``piece length`` from existing torrent
+
+        If `path` is a directory, it is searched recursively for a matching
+        ``.torrent`` file. Otherwise, it is :meth:`read` as a ``.torrent`` file.
+
+        `path` may also be an iterable, in which case each item is treated as
+        described above until a match is found.
+
+        A matching torrent is found by searching for a torrent with the same
+        :attr:`name` and :attr:`files` (file size must also match).  If a match
+        is found, compare three piece hashes per file to reduce the risk of a
+        false positive.
+
+        .. warning:: This should almost always work in practice, but a false
+            positive match is theoretically possible, and there is no way to
+            avoid that.
+
+        .. note:: This will likely change :attr:`infohash` and turn this
+            instance into a new torrent.
+
+        :param path: Path to torrent file or directory or iterable of
+            file/directory paths
+        :param callable callback: Callable to report progress and/or abort
+
+            `callback` must accept 6 positional arguments:
+
+                1. Torrent instance (:class:`Torrent`)
+                2. Torrent file path (:class:`str`) or ``None``
+                3. Number of checked torrent files (:class:`int`)
+                4. Total number of torrent files (:class:`int`)
+                5. Status indicator:
+                     ``True``  - Torrent file is a match
+                     ``False`` - Torrent file is not a match
+                     ``None`` - File trees are identical and we are now
+                                comparing a few piece hashes
+                6. Exception (:class:`TorfError`) or ``None``
+
+            If `callback` returns anything that is not ``None``, stop searching.
+        :param float interval: Minimum number of seconds between calls to
+            `callback`; if 0, `callback` is called for each torrent file;
+            `callback` is always called if `exception` is not ``None``
+
+        :raises ReadError: if reading a torrent file fails
+        :raises BdecodeError: if parsing a torrent file fails
+        :raises MetainfoError: if a torrent file contains invalid or
+            insufficient metadata
+
+        :return: `True` if a matching torrent was found, `False` otherwise
+        """
+        if not self.path:
+            raise RuntimeError('reuse() called with no path specified')
+        elif not self.files:
+            raise RuntimeError('reuse() called while file list is empty')
+
+        if isinstance(path, (str, pathlib.PurePath)):
+            paths = [path]
+        elif isinstance(path, abc.Iterable):
+            paths = tuple(path)
+        else:
+            raise ValueError(f'Invalid path argument: {path!r}')
+
+        torrent_file_items = reuse.find_torrent_files(*paths, max_file_size=self.MAX_TORRENT_FILE_SIZE)
+        maybe_call_callback = reuse.ReuseCallback(
+            callback=callback,
+            interval=interval,
+            torrent=self,
+            torrent_files_total=torrent_file_items.total,
+        )
+
+        for candidate_path, files_done, exception in torrent_file_items:
+            try:
+                if candidate_path:
+                    candidate = Torrent.read(candidate_path)
+                elif exception:
+                    raise exception
+                else:
+                    raise RuntimeError('Both candidate_path and exception are None?!')
+
+            except (error.ReadError, error.BdecodeError, error.MetainfoError) as e:
+                cancelled = maybe_call_callback(candidate_path, files_done, False, e)
+                if cancelled is not None:
+                    break
+
+            else:
+                assert exception is None
+                if reuse.is_file_match(self, candidate):
+                    cancelled = maybe_call_callback(candidate_path, files_done, None, exception)
+                    if cancelled is not None:
+                        break
+
+                    if reuse.is_content_match(self, candidate):
+                        maybe_call_callback(candidate_path, files_done, True, exception)
+                        reuse.copy(candidate, self)
+                        return True
+                    else:
+                        cancelled = maybe_call_callback(candidate_path, files_done, False, exception)
+                        if cancelled is not None:
+                            break
+
+                else:
+                    cancelled = maybe_call_callback(candidate_path, files_done, False, exception)
+                    if cancelled is not None:
+                        break
+
+        return False
+
     def __repr__(self):
         sig = inspect.signature(self.__init__)
         args = []
